/*
 noniso.cpp - replacements for non-ISO functions used by Arduino core
 Copyright � 2016 Ivan Grokhotkov
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 */

#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <stdint.h>
#include <math.h>
#include "stdlib_noniso.h"

void reverse( char *begin, char *end )
{
    char *is = begin;
    char *ie = end - 1;
    while( is < ie )
    {
        char tmp = *ie;
        *ie      = *is;
        *is      = tmp;
        ++is;
        --ie;
    }
}

char *ltoa( long value, char *result, int base )
{
    if( base < 2 || base > 16 )
    {
        *result = 0;
        return result;
    }

    char *out      = result;
    long  quotient = abs( value );

    do
    {
        const long tmp = quotient / base;
        *out           = "0123456789abcdef"[quotient - ( tmp * base )];
        ++out;
        quotient = tmp;
    } while( quotient );

    // Apply negative sign
    if( value < 0 ) *out++ = '-';

    reverse( result, out );
    *out = 0;
    return result;
}

char *ultoa( unsigned long value, char *result, int base )
{
    if( base < 2 || base > 16 )
    {
        *result = 0;
        return result;
    }

    char *        out      = result;
    unsigned long quotient = value;

    do
    {
        const unsigned long tmp = quotient / base;
        *out                    = "0123456789abcdef"[quotient - ( tmp * base )];
        ++out;
        quotient = tmp;
    } while( quotient );

    reverse( result, out );
    *out = 0;
    return result;
}

char *dtostrf( double number, signed char width, unsigned char prec, char *s )
{
    bool negative = false;

    if( isnan( number ) )
    {
        strcpy( s, "nan" );
        return s;
    }
    if( isinf( number ) )
    {
        strcpy( s, "inf" );
        return s;
    }

    char *out = s;

    int fillme = width;    // how many cells to fill for the integer part
    if( prec > 0 ) { fillme -= ( prec + 1 ); }

    // Handle negative numbers
    if( number < 0.0 )
    {
        negative = true;
        fillme--;
        number = -number;
    }

    // Round correctly so that print(1.999, 2) prints as "2.00"
    // I optimized out most of the divisions
    double rounding = 2.0;
    for( uint8_t i = 0; i < prec; ++i ) rounding *= 10.0;
    rounding = 1.0 / rounding;

    number += rounding;

    // Figure out how big our number really is
    double tenpow     = 1.0;
    int    digitcount = 1;
    while( number >= 10.0 * tenpow )
    {
        tenpow *= 10.0;
        digitcount++;
    }

    number /= tenpow;
    fillme -= digitcount;

    // Pad unused cells with spaces
    while( fillme-- > 0 ) { *out++ = ' '; }

    // Handle negative sign
    if( negative ) *out++ = '-';

    // Print the digits, and if necessary, the decimal point
    digitcount += prec;
    int8_t digit = 0;
    while( digitcount-- > 0 )
    {
        digit = ( int8_t )number;
        if( digit > 9 ) digit = 9;    // insurance
        *out++ = ( char )( '0' | digit );
        if( ( digitcount == prec ) && ( prec > 0 ) ) { *out++ = '.'; }
        number -= digit;
        number *= 10.0;
    }

    // make sure the string is terminated
    *out = 0;
    return s;
}

char *utoa( unsigned value, char *result, int base )
{
    if( base < 2 || base > 16 )
    {
        *result = 0;
        return result;
    }

    char *   out      = result;
    unsigned quotient = value;

    do
    {
        const unsigned tmp = quotient / base;
        *out               = "0123456789abcdef"[quotient - ( tmp * base )];
        ++out;
        quotient = tmp;
    } while( quotient );

    reverse( result, out );
    *out = 0;
    return result;
}

char *itoa( int value, char *result, int base )
{
    if( base < 2 || base > 16 )
    {
        *result = 0;
        return result;
    }

    char *out      = result;
    int   quotient = abs( value );

    do
    {
        const int tmp = quotient / base;
        *out          = "0123456789abcdef"[quotient - ( tmp * base )];
        ++out;
        quotient = tmp;
    } while( quotient );

    // Apply negative sign
    if( value < 0 ) *out++ = '-';

    reverse( result, out );
    *out = 0;
    return result;
}
